"""Base class for all adapters."""

from __future__ import annotations

import datetime
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Sequence

from icalendar.prop import vDDDTypes

from recurring_ical_events.util import (
    cached_property,
    make_comparable,
    time_span_contains_event,
    to_recurrence_ids,
)

if TYPE_CHECKING:
    from icalendar import Alarm
    from icalendar.cal import Component

    from recurring_ical_events.series import Series
    from recurring_ical_events.types import UID, RecurrenceIDs, Time


class ComponentAdapter(ABC):
    """A unified interface to work with icalendar components."""

    ATTRIBUTES_TO_DELETE_ON_COPY = ["RRULE", "RDATE", "EXDATE"]

    @staticmethod
    @abstractmethod
    def component_name() -> str:
        """The icalendar component name."""

    def __init__(self, component: Component):
        """Create a new adapter."""
        self._component = component

    @property
    def alarms(self) -> list[Alarm]:
        """The alarms in this component."""
        return self._component.walk("VALARM")

    @property
    def end_property(self) -> str | None:
        """The name of the end property."""
        return None

    @property
    def start(self) -> Time:
        """The start time."""
        return self.span[0]

    @property
    def end(self) -> Time:
        """The end time."""
        return self.span[1]

    @cached_property
    def span(self):
        """Return (start, end)."""
        start, end = make_comparable((self.raw_start, self.raw_end))
        if start > end:
            return end, start
        return start, end

    @property
    @abstractmethod
    def raw_start(self):
        """Return the start property of the component."""

    @property
    @abstractmethod
    def raw_end(self):
        """Return the start property of the component."""

    @property
    def uid(self) -> UID:
        """The UID of a component.

        UID is required by RFC5545.
        If the UID is absent, we use the Python ID.
        """
        return self._component.get("UID", str(id(self._component)))

    @classmethod
    def collect_series_from(
        cls, source: Component, suppress_errors: tuple[Exception]
    ) -> Sequence[Series]:
        """Collect all components for this adapter.

        This is a shortcut.
        """
        from recurring_ical_events.selection.name import ComponentsWithName

        return ComponentsWithName(cls.component_name(), cls).collect_series_from(
            source, suppress_errors
        )

    def as_component(
        self,
        start: Optional[Time] = None,
        stop: Optional[Time] = None,
        keep_recurrence_attributes: bool = True,  # noqa: FBT001
    ):
        """Create a shallow copy of the source event and modify some attributes."""
        copied_component = self._component.copy()
        copied_component["DTSTART"] = vDDDTypes(self.start if start is None else start)
        copied_component.pop("DURATION", None)  # remove duplication in event length
        if self.end_property is not None:
            copied_component[self.end_property] = vDDDTypes(
                self.end if stop is None else stop
            )
        if not keep_recurrence_attributes:
            for attribute in self.ATTRIBUTES_TO_DELETE_ON_COPY:
                if attribute in copied_component:
                    del copied_component[attribute]
        for subcomponent in self._component.subcomponents:
            copied_component.add_component(subcomponent)
        if "RECURRENCE-ID" not in copied_component:
            copied_component["RECURRENCE-ID"] = vDDDTypes(
                copied_component["DTSTART"].dt
            )
        return copied_component

    @cached_property
    def recurrence_ids(self) -> RecurrenceIDs:
        """The recurrence ids of the component that might be used to identify it."""
        recurrence_id = self._component.get("RECURRENCE-ID")
        if recurrence_id is None:
            return ()
        return to_recurrence_ids(recurrence_id.dt)

    @cached_property
    def this_and_future(self) -> bool:
        """The recurrence ids has a thisand future range property"""
        recurrence_id = self._component.get("RECURRENCE-ID")
        if recurrence_id is None:
            return False
        if "RANGE" in recurrence_id.params:
            return recurrence_id.params["RANGE"] == "THISANDFUTURE"
        return False

    def is_modification(self) -> bool:
        """Whether the adapter is a modification."""
        return bool(self.recurrence_ids)

    @cached_property
    def sequence(self) -> int:
        """The sequence in the history of modification.

        The sequence is negative if none was found.
        """
        return self._component.get("SEQUENCE", -1)

    def __repr__(self) -> str:
        """Debug representation with more info."""
        return (
            f"<{self.__class__.__name__} UID={self.uid} start={self.start} "
            f"recurrence_ids={self.recurrence_ids} sequence={self.sequence} "
            f"end={self.end}>"
        )

    @cached_property
    def exdates(self) -> list[Time]:
        """A list of exdates."""
        result: list[Time] = []
        exdates = self._component.get("EXDATE", [])
        for exdates in (exdates,) if not isinstance(exdates, list) else exdates:
            result.extend(exdate.dt for exdate in exdates.dts)
        return result

    @cached_property
    def rrules(self) -> set[str]:
        """A list of rrules of this component."""
        rules = self._component.get("RRULE", None)
        if not rules:
            return set()
        return {
            rrule.to_ical().decode()
            for rrule in (rules if isinstance(rules, list) else [rules])
        }

    @cached_property
    def rdates(self) -> list[Time, tuple[Time, Time]]:
        """A list of rdates, possibly a period."""
        rdates = self._component.get("RDATE", [])
        result = []
        for rdates in (rdates,) if not isinstance(rdates, list) else rdates:
            result.extend(rdate.dt for rdate in rdates.dts)
        return result

    @cached_property
    def duration(self) -> datetime.timedelta:
        """The duration of the component."""
        return self.end - self.start

    def is_in_span(self, span_start: Time, span_stop: Time) -> bool:
        """Return whether the component is in the span."""
        return time_span_contains_event(span_start, span_stop, self.start, self.end)

    @cached_property
    def extend_query_span_by(self) -> tuple[datetime.timedelta, datetime.timedelta]:
        """Calculate how much we extend the query span.

        If an event is long, we need to extend the query span by the event's duration.
        If an event has moved, we need to make sure that that is included, too.

        This is so that the RECURRENCE-ID falls within the modified span.
        Imagine if the span is exactly a second. How much would we need to query
        forward and backward to capture the recurrence id?

        Returns two positive spans: (subtract_from_start, add_to_stop)
        """
        subtract_from_start = self.duration
        add_to_stop = datetime.timedelta(0)
        recurrence_id_prop = self._component.get("RECURRENCE-ID")
        if recurrence_id_prop:
            start, end, recurrence_id = make_comparable(
                (self.start, self.end, recurrence_id_prop.dt)
            )
            if start < recurrence_id:
                add_to_stop = recurrence_id - start
            if start > recurrence_id:
                subtract_from_start = end - recurrence_id
        return subtract_from_start, add_to_stop

    @cached_property
    def move_recurrences_by(self) -> datetime.timedelta:
        """Occurrences of this component should be moved by this amount.

        Usually, the occurrence starts at the new start time.
        However, if we have a RANGE=THISANDFUTURE, we need to move the occurrence.

        RFC 5545:

            When the given recurrence instance is
            rescheduled, all subsequent instances are also rescheduled by the
            same time difference.  For instance, if the given recurrence
            instance is rescheduled to start 2 hours later, then all
            subsequent instances are also rescheduled 2 hours later.
            Similarly, if the duration of the given recurrence instance is
            modified, then all subsequence instances are also modified to have
            this same duration.
        """
        if self.this_and_future:
            recurrence_id_prop = self._component.get("RECURRENCE-ID")
            assert recurrence_id_prop, "RANGE=THISANDFUTURE implies RECURRENCE-ID."
            start, recurrence_id = make_comparable((self.start, recurrence_id_prop.dt))
            return start - recurrence_id
        return datetime.timedelta(0)


__all__ = ["ComponentAdapter"]



"""Constants for recurring_ical_events."""

import datetime
import re
from pathlib import Path

# The minimum value accepted as date (pytz + zoneinfo)
DATE_MIN = (1970, 1, 1)
DATE_MIN_DT = datetime.date(*DATE_MIN)
# The maximum value accepted as date (pytz + zoneinfo)
DATE_MAX = (2038, 1, 1)
DATE_MAX_DT = datetime.date(*DATE_MAX)

# the location of this file
HERE = Path(__file__).parent

# the directory with all example calendars
CALENDARS = HERE / "test" / "calendars"

NEGATIVE_RRULE_COUNT_REGEX = re.compile(r"COUNT=-\d+;?")

__all__ = [
    "CALENDARS",
    "DATE_MAX",
    "DATE_MAX_DT",
    "DATE_MIN",
    "DATE_MIN_DT",
    "NEGATIVE_RRULE_COUNT_REGEX",
]



"""Functionality for smaller examples."""

import icalendar

from recurring_ical_events.constants import CALENDARS


def example_calendar(name: str = "") -> icalendar.Calendar:
    """Return an example calendar.

    Args:
        name (str): The name of the example file.

    Returns:
        icalendar.cal.Calendar: The parsed calendar example.
    """
    if not name.endswith(".ics"):
        name += ".ics"
    path = CALENDARS / name
    try:
        return icalendar.Calendar.from_ical(path.read_bytes())
    except FileNotFoundError:
        raise ValueError(  # noqa: B904
            f"File {name!r} not found. "
            f"Use one of {', '.join(p.name for p in CALENDARS.glob('*.ics'))!r}."
        )


__all__ = ["example_calendar"]



"""Adapter for VJOURNAL."""

from recurring_ical_events.adapters.component import ComponentAdapter
from recurring_ical_events.constants import DATE_MIN_DT
from recurring_ical_events.types import Time
from recurring_ical_events.util import cached_property


class JournalAdapter(ComponentAdapter):
    """Apdater for journal entries."""

    @staticmethod
    def component_name() -> str:
        """The icalendar component name."""
        return "VJOURNAL"

    @property
    def end_property(self) -> None:
        """There is no end property"""

    @property
    def raw_start(self) -> Time:
        """Return DTSTART if it set, do not panic if it's not set."""
        ## according to the specification, DTSTART in a VJOURNAL is optional
        dtstart = self._component.get("DTSTART")
        if dtstart is not None:
            return dtstart.dt
        return DATE_MIN_DT

    @cached_property
    def raw_end(self) -> Time:
        """The end time is the same as the start."""
        ## VJOURNAL cannot have a DTEND.  We should consider a VJOURNAL to
        ## describe one day if DTSTART is a date, and we can probably
        ## consider it to have zero duration if a timestamp is given.
        return self.raw_start


__all__ = ["JournalAdapter"]



"""Series calculation for alarms."""

import datetime
from collections import defaultdict
from typing import Generator

from dateutil.rrule import rruleset
from icalendar import Alarm

from recurring_ical_events.adapters.component import ComponentAdapter
from recurring_ical_events.occurrence import AlarmOccurrence, Occurrence
from recurring_ical_events.series.rrule import Series
from recurring_ical_events.types import Time
from recurring_ical_events.util import convert_to_datetime


class AbsoluteAlarmSeries:
    """A series of absolute alarms."""

    tzinfo = datetime.timezone.utc

    def __init__(self):
        """Create a new series of absolute alarms."""
        self.times = rruleset(cache=True)
        self.times2occurence: dict[datetime.datetime, list[Occurrence]] = defaultdict(
            list
        )

    def add(self, alarm: Alarm, parent: ComponentAdapter):
        """Add an absolute alarm with a parent component."""
        trigger = alarm.TRIGGER
        self._add(trigger, alarm, parent)
        for _ in range(alarm.REPEAT):
            trigger += alarm.DURATION
            self._add(trigger, alarm, parent)

    def _add(self, dt: datetime.datetime, alarm: Alarm, parent: ComponentAdapter):
        """Add an alarm at a specific time."""
        self.times.rdate(dt)
        self.times2occurence[dt].append(self.occurrence(dt, alarm, parent))

    def between(
        self, span_start: Time, span_stop: Time
    ) -> Generator[Occurrence, None, None]:
        """Components between the start (inclusive) and end (exclusive).

        The result does not need to be ordered.
        """
        span_start_dt = convert_to_datetime(span_start, self.tzinfo)
        span_stop_dt = convert_to_datetime(span_stop, self.tzinfo)
        for dt in self.times.between(span_start_dt, span_stop_dt, inc=True):
            for occurrence in self.times2occurence[dt]:
                if occurrence.is_in_span(span_start_dt, span_stop_dt):
                    yield occurrence

    def occurrence(
        self, dt: datetime.datetime, alarm: Alarm, parent: ComponentAdapter
    ) -> Occurrence:
        """Create a new occurrence."""
        return AlarmOccurrence(dt, alarm, parent)

    def is_empty(self) -> bool:
        """Whether this series is empty."""
        return not self.times2occurence


class AlarmSeriesRelativeToStart:
    """A series of alarms relative to the start of a component."""

    def __init__(self, alarm: Alarm, series: Series) -> None:
        """Create a series of alarms that are relative to the start of a series."""
        self._alarm = alarm
        self._series = series
        self._offsets: list[datetime.timedelta] = [alarm.TRIGGER]
        for _ in range(alarm.REPEAT):
            self._offsets.append(self._offsets[-1] + alarm.DURATION)

    def between(
        self, span_start: Time, span_stop: Time
    ) -> Generator[Occurrence, None, None]:
        """Components between the start (inclusive) and end (exclusive).

        The result does not need to be ordered.
        """
        # TODO: Reduce time span to reduce occurrences
        for offset in self._offsets:
            # If we are before the event start (negative offset),
            # we have to add the time span to request the event later.
            for parent in self._series.between(span_start - offset, span_stop - offset):
                if parent.has_alarm(self._alarm):
                    occurrence = self.occurrence(offset, self._alarm, parent)
                    if occurrence.is_in_span(span_start, span_stop):
                        yield occurrence

    def occurrence(
        self, offset: datetime.timedelta, alarm: Alarm, parent: Occurrence
    ) -> Occurrence:
        """Create a new occurrence."""
        return AlarmOccurrence(offset + parent.start, alarm, parent)

    def __repr__(self) -> str:
        """repr()"""
        return (
            f"<{self.__class__.__name__} "
            f"of {self._alarm} in {self._series} "
            f"with offsets {', '.join(map(str, self._offsets))}>"
        )


class AlarmSeriesRelativeToEnd(AlarmSeriesRelativeToStart):
    """A series of alarms relative to the start of a component."""

    def between(self, span_start, span_stop):
        """Components between the start (inclusive) and end (exclusive).

        The result does not need to be ordered.
        """
        # The end is exclusive. We must adjust the timespan to include it.
        return super().between(span_start - datetime.timedelta(seconds=1), span_stop)

    def occurrence(
        self, offset: datetime.timedelta, alarm: Alarm, parent: Occurrence
    ) -> Occurrence:
        """Create a new occurrence."""
        return AlarmOccurrence(offset + parent.end, alarm, parent)


__all__ = [
    "AbsoluteAlarmSeries",
    "AlarmSeriesRelativeToEnd",
    "AlarmSeriesRelativeToStart",
]



"""All the errors."""

from recurring_ical_events.types import Time


class InvalidCalendar(ValueError):
    """Exception thrown for bad icalendar content."""

    def __init__(self, message: str):
        """Create a new error with a message."""
        self._message = message
        super().__init__(self.message)

    @property
    def message(self) -> str:
        """The error message."""
        return self._message


class PeriodEndBeforeStart(InvalidCalendar):
    """An event or component starts before it ends."""

    def __init__(self, message: str, start: Time, end: Time):
        """Create a new PeriodEndBeforeStart error."""
        super().__init__(message)
        self._start = start
        self._end = end

    @property
    def start(self) -> Time:
        """The start of the component's period."""
        return self._start

    @property
    def end(self) -> Time:
        """The end of the component's period."""
        return self._end


class BadRuleStringFormat(InvalidCalendar):
    """An iCal rule string is badly formatted."""

    def __init__(self, message: str, rule: str):
        """Create an error with a bad rule string."""
        super().__init__(message + ": " + rule)
        self._rule = rule

    @property
    def rule(self) -> str:
        """The malformed rule string"""
        return self._rule


__all__ = ["BadRuleStringFormat", "InvalidCalendar", "PeriodEndBeforeStart"]



"""Occurrences of events and other components."""

from __future__ import annotations

from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, NamedTuple, Optional

from icalendar import Alarm

from recurring_ical_events.adapters.component import ComponentAdapter
from recurring_ical_events.util import (
    cached_property,
    make_comparable,
    time_span_contains_event,
)

if TYPE_CHECKING:
    from icalendar import Alarm
    from icalendar.cal import Component

    from recurring_ical_events.adapters.component import ComponentAdapter
    from recurring_ical_events.types import UID, RecurrenceIDs, Time


class OccurrenceID(NamedTuple):
    """The ID of a component's occurrence to identify it clearly.

    Attributes:
        name: The name of the component, e.g. "VEVENT"
        uid: The UID of the component.
        recurrence_id: The Recurrence-ID of the component in UTC but without tzinfo.
        start: The start of the component

    """

    name: str
    uid: UID
    recurrence_id: Optional[Time]
    start: Time

    def to_string(self) -> str:
        """Return a string representation of this id."""
        return "#".join(
            [
                self.name,
                self.recurrence_id.isoformat() if self.recurrence_id else "",
                self.start.isoformat(),
                self.uid,
            ]
        )

    @staticmethod
    def _dt_from_string(iso_string: str) -> Time:
        """Create a datetime from the string representation."""
        if len(iso_string) == 10:
            return date.fromisoformat(iso_string)
        return datetime.fromisoformat(iso_string)

    @classmethod
    def from_string(cls, string_id: str) -> OccurrenceID:
        """Parse a string and return the component id."""
        name, recurrence_id, start, uid = string_id.split("#", 3)
        return cls(
            name,
            uid,
            cls._dt_from_string(recurrence_id) if recurrence_id else None,
            cls._dt_from_string(start),
        )

    @classmethod
    def from_occurrence(
        cls, name: str, uid: str, recurrence_ids: RecurrenceIDs, start: Time
    ):
        """Create a new OccurrenceID from the given values.

        Args:
            name: The component name.
            uid: The UID string.
            recurrence_ids: The recurrence ID tuple.
                This is expected as UTC with tzinfo being None.
            start: start time of the component either with or without timezone
        """
        return cls(name, uid, recurrence_ids[0] if recurrence_ids else None, start)


class Occurrence:
    """A repetition of an event."""

    def __init__(
        self,
        adapter: ComponentAdapter,
        start: Time | None = None,
        end: Time | None | timedelta = None,
        sequence: int = -1,
    ):
        """Create an event repetition.

        - source - the icalendar Event
        - start - the start date/datetime to replace
        - stop - the end date/datetime to replace
        - sequence - if positive or 0, this sets the SEQUENCE property
        """
        self._adapter = adapter
        self.start = adapter.start if start is None else start
        self.end = adapter.end if end is None else end
        self.sequence = sequence

    def as_component(self, keep_recurrence_attributes: bool) -> Component:  # noqa: FBT001
        """Create a shallow copy of the source component and modify some attributes."""
        component = self._adapter.as_component(
            self.start, self.end, keep_recurrence_attributes
        )
        if self.sequence >= 0:
            component["SEQUENCE"] = self.sequence
        return component

    def is_in_span(self, span_start: Time, span_stop: Time) -> bool:
        """Return whether the component is in the span."""
        return time_span_contains_event(span_start, span_stop, self.start, self.end)

    def __lt__(self, other: Occurrence) -> bool:
        """Compare two occurrences for sorting.

        See https://stackoverflow.com/a/4010558/1320237
        """
        self_start, other_start = make_comparable((self.start, other.start))
        return self_start < other_start

    @cached_property
    def id(self) -> OccurrenceID:
        """The id of the component."""
        return OccurrenceID.from_occurrence(
            self._adapter.component_name(),
            self._adapter.uid,
            self._adapter.recurrence_ids,
            self.start,
        )

    def __hash__(self) -> int:
        """Hash this for an occurrence."""
        return hash(self.id)

    def __eq__(self, other: Occurrence) -> bool:
        """self == other"""
        return self.id == other.id

    def component_name(self) -> str:
        """The name of this component."""
        return self._adapter.component_name()

    @property
    def uid(self) -> str:
        """The UID of this occurrence."""
        return self._adapter.uid

    def has_alarm(self, alarm: Alarm) -> bool:
        """Wether this alarm is in this occurrence."""
        return alarm in self._adapter.alarms

    @property
    def recurrence_ids(self) -> RecurrenceIDs:
        """The recurrence ids."""
        return self._adapter.recurrence_ids


class AlarmOccurrence(Occurrence):
    """Adapter for absolute alarms."""

    def __init__(
        self,
        trigger: datetime,
        alarm: Alarm,
        parent: ComponentAdapter | Occurrence,
    ) -> None:
        super().__init__(alarm, trigger, trigger)
        self.parent = parent
        self.alarm = alarm

    def as_component(self, keep_recurrence_attributes):
        """Return the alarm's parent as a modified component."""
        parent = self.parent.as_component(
            keep_recurrence_attributes=keep_recurrence_attributes
        )
        alarm_once = self.alarm.copy()
        alarm_once.TRIGGER = self.start
        alarm_once.REPEAT = 0
        parent.subcomponents = [alarm_once]
        return parent

    @cached_property
    def id(self) -> OccurrenceID:
        """The id of the component."""
        return OccurrenceID.from_occurrence(
            self.parent.component_name(),
            self.parent.uid,
            self.parent.recurrence_ids,
            self.start,
        )

    def __repr__(self) -> str:
        """repr(self)"""
        return (
            f"<{self.__class__.__name__} at {self.start} of"
            f" {self.alarm} in {self.parent}"
        )


__all__ = [
    "AlarmOccurrence",
    "Occurrence",
    "OccurrenceID",
]



"""Adapter for VEVENT."""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from recurring_ical_events.adapters.component import ComponentAdapter
from recurring_ical_events.util import (
    convert_to_datetime,
    is_date,
    normalize_pytz,
)

if TYPE_CHECKING:
    from recurring_ical_events.types import Time


class EventAdapter(ComponentAdapter):
    """An icalendar event adapter."""

    @staticmethod
    def component_name() -> str:
        """The icalendar component name."""
        return "VEVENT"

    @property
    def end_property(self) -> str:
        """DTEND"""
        return "DTEND"

    @property
    def raw_start(self) -> Time:
        """Return DTSTART"""
        # Arguably, it may be considered a feature that this breaks
        # if no DTSTART is set
        return self._component["DTSTART"].dt

    @property
    def raw_end(self) -> Time:
        """Yield DTEND or calculate the end of the event based on
        DTSTART and DURATION.
        """
        ## an even may have DTEND or DURATION, but not both
        end = self._component.get("DTEND")
        if end is not None:
            return end.dt
        duration = self._component.get("DURATION")
        if duration is not None:
            start = self._component["DTSTART"].dt
            if duration.dt.seconds != 0 and is_date(start):
                start = convert_to_datetime(start, None)
            return normalize_pytz(start + duration.dt)
        start = self._component["DTSTART"].dt
        if is_date(start):
            return start + datetime.timedelta(days=1)
        return start


__all__ = ["EventAdapter"]
